"use strict";
// Copyright (c) 2018, Shellyl_N and Authors
// license: ISC
// https://github.com/shellyln
Object.defineProperty(exports, "__esModule", { value: true });
const ast_1 = require("./ast");
const evaluate_1 = require("./evaluate");
const errors_1 = require("./errors");
const helpers_1 = require("./compile.ops/helpers");
const compile_ops_1 = require("./compile.ops");
// tslint:disable-next-line:variable-name
function compileCore(state, formalArgs, lastIsSpread, fnBody, _$_vars) {
    function makeScope(scoped) {
        const varNamesCopy = new Map();
        for (const x of ctx.varNames.entries()) {
            varNamesCopy.set(x[0], x[1]);
        }
        scoped();
        ctx.varNames = varNamesCopy;
    }
    function compileValue(b) {
        let compFnBody = '';
        if (b === null) {
            compFnBody += '(null)';
        }
        else if (b === void 0) {
            compFnBody += '(void 0)';
        }
        else
            switch (typeof b) {
                case 'boolean':
                case 'number':
                    compFnBody += `(${String(b)})`;
                    break;
                case 'object':
                    if (ast_1.isSymbol(b)) {
                        const sym = b;
                        errors_1.checkUnsafeVarNamesEx('(compiler)compileValue', ctx.varNames, sym.symbol);
                        if (ctx.varNames.has(sym.symbol)) {
                            compFnBody += `(${ctx.varNames.get(sym.symbol)})`;
                        }
                        else {
                            _$_vars[ctx.varsCount] = sym.symbol;
                            compFnBody += `(_$_vars[${String(ctx.varsCount++)}])`;
                        }
                        break;
                    }
                    else { }
                // FALL_THRU
                default:
                    _$_vars[ctx.varsCount] = b;
                    compFnBody += `(_$_vars[${String(ctx.varsCount++)}])`;
                    break;
            }
        return compFnBody;
    }
    function compileToken(body, i) {
        let compFnBody = '';
        const b = body[i];
        if (Array.isArray(b)) {
            if (0 < b.length) {
                const r = helpers_1.applyMacros(state, b);
                if (Array.isArray(r)) {
                    if (0 < r.length) {
                        if (ast_1.isSymbol(r[0])) {
                            const sym = r[0];
                            const args = r.slice(1);
                            errors_1.checkUnsafeVarNamesEx('(compiler)compileToken', ops, sym.symbol);
                            if (ops.has(sym.symbol)) {
                                compFnBody += ops.get(sym.symbol)(r, args);
                            }
                            else {
                                if (sym.symbol === state.config.reservedNames.spread) {
                                    compFnBody += `...(${args.map((x) => compileToken([helpers_1.stripQuoteOrPass(state, x)], 0)).join(',')})`;
                                }
                                else if (state.funcMap.has(sym.symbol)) {
                                    _$_vars[ctx.varsCount] = state.funcMap.get(sym.symbol).fn(state, '');
                                    compFnBody += `((_$_vars[${String(ctx.varsCount++)}])(${args.map((x) => compileToken([helpers_1.stripQuoteOrPass(state, x)], 0)).join(',')}))`;
                                }
                                else if (ctx.varNames.has(sym.symbol)) {
                                    compFnBody += `(${String(ctx.varNames.get(sym.symbol))})(${args.map((x) => compileToken([helpers_1.stripQuoteOrPass(state, x)], 0)).join(',')})`;
                                }
                                else {
                                    _$_vars[ctx.varsCount] = helpers_1.resolveValueSymbol_dynamic(state, sym.symbol);
                                    compFnBody += `((_$_vars[${String(ctx.varsCount++)}])()(${args.map((x) => compileToken([helpers_1.stripQuoteOrPass(state, x)], 0)).join(',')}))`;
                                }
                            }
                        }
                        else {
                            switch (typeof r[0]) {
                                case 'function':
                                    _$_vars[ctx.varsCount] = r[0];
                                    compFnBody += `((_$_vars[${String(ctx.varsCount++)}])(${r.slice(1).map((x, idx, arr) => compileToken([helpers_1.stripQuoteOrPass(state, x)], 0)).join(',')}))`;
                                    break;
                                default:
                                    throw new Error(`[SX] compileToken: First item of list is not a function: ${JSON.stringify(r[0])}.`);
                            }
                        }
                    }
                    else {
                        compFnBody += '([])';
                    }
                }
                else {
                    compFnBody += compileValue(b);
                }
            }
            else {
                compFnBody += '([])';
            }
        }
        else {
            compFnBody += compileValue(b);
        }
        return compFnBody;
    }
    const ctx = {
        _$_vars,
        varsCount: 1,
        varNames: new Map(),
        varNamesCount: 0,
        varDefs: 'var x0;',
        ops: new Map(),
        makeScope,
        compileToken,
    };
    compile_ops_1.registerOperators(state, ctx);
    const ops = ctx.ops;
    if (state.config.enableTailCallOptimization) {
        fnBody = evaluate_1.optimizeTailCall(state, formalArgs, fnBody);
    }
    // set 'this' object
    ctx.varNames.set(state.config.reservedNames.thiz, `(this===(Function('return this')())?null:(this===void 0?null:this))`);
    const capturedScopes = evaluate_1.getCapturedScopes(state);
    if (capturedScopes) {
        for (const x in capturedScopes) {
            if (Object.prototype.hasOwnProperty.call(capturedScopes, x)) {
                _$_vars[ctx.varsCount++] = x;
                _$_vars[ctx.varsCount] = capturedScopes[x];
                ctx.varNames.set(x, `(_$_vars[${String(ctx.varsCount)}][_$_vars[${String(ctx.varsCount - 1)}]])`);
                ctx.varsCount++;
            }
        }
    }
    const compFormalArgs = `${formalArgs.map((x, i) => {
        ctx.varNames.set(formalArgs[i].symbol, 'a' + i);
        return `${(lastIsSpread && i === formalArgs.length - 1) ? '...' : ''}a${i}`;
    }).join(',')}`;
    const compFnBodyRoot = `return(${fnBody.map((x, i) => compileToken(fnBody, i)).join(',')})`;
    return `(function(${compFormalArgs}){"strict";${ctx.varDefs}${compFnBodyRoot}})`;
}
// tslint:disable-next-line:variable-name
function evalCompiledLambda(_$_state, _$_vars, code) {
    // tslint:disable-next-line:no-eval
    _$_vars[0] = eval(code);
    return _$_vars[0];
}
function compileLambda(state, formalArgs, lastIsSpread, fnBody) {
    // tslint:disable-next-line:variable-name
    const _$_vars = [];
    return evalCompiledLambda(state, _$_vars, compileCore(state, formalArgs, lastIsSpread, fnBody, _$_vars));
}
exports.compileLambda = compileLambda;
//# sourceMappingURL=compile.js.map