"use strict";
// Copyright (c) 2018, Shellyl_N and Authors
// license: ISC
// https://github.com/shellyln
Object.defineProperty(exports, "__esModule", { value: true });
const types_1 = require("./types");
const consts_1 = require("./consts");
function setEvaluationCount(state) {
    state.evalCount++;
    if (state.config.maxEvalCount && state.config.maxEvalCount < state.evalCount) {
        throw new types_1.MaxEvaluationCountError();
    }
}
exports.setEvaluationCount = setEvaluationCount;
function checkParamsLength(name, args, min, max) {
    if (args.length < min) {
        throw new Error(`[SX] ${name}: Invalid argument length: expected: ${min} / args: ${args.length}.`);
    }
    if (max && max < args.length) {
        throw new Error(`[SX] ${name}: Invalid argument length: expected: ${max} / args: ${args.length}.`);
    }
    return args;
}
exports.checkParamsLength = checkParamsLength;
const objConstructor = ({}).constructor; // NOTE: objConstructor            === Object
const funConstructor = Function; // NOTE: ({}).toString.constructor === Function
function checkUnsafeVarNames(name, varName) {
    if (varName === '__proto__' ||
        varName === '__defineGetter__' || varName === '__defineSetter__' ||
        varName === '__lookupGetter__' || varName === '__lookupSetter__') {
        throw new Error(`[SX] ${name}: Invalid var name ${varName}.`);
    }
    if (varName === 'prototype' || varName === 'constructor') {
        throw new Error(`[SX] ${name}: Invalid var name ${varName}.`);
    }
    if (objConstructor.hasOwnProperty(varName)) {
        throw new Error(`[SX] ${name}: Invalid var name ${varName}.`);
    }
    if (varName === 'call' || varName === 'arguments' || varName === 'caller') {
        // NOTE: arguments, caller are not accessible in strict mode
        throw new Error(`[SX] ${name}: Invalid var name ${varName}.`);
    }
    return varName;
}
exports.checkUnsafeVarNames = checkUnsafeVarNames;
function checkUnsafeVarNamesEx(name, target, varName) {
    if (target === consts_1.globalObj ||
        varName === '__proto__' ||
        varName === '__defineGetter__' || varName === '__defineSetter__' ||
        varName === '__lookupGetter__' || varName === '__lookupSetter__') {
        throw new Error(`[SX] ${name}: Invalid var name ${varName}.`);
    }
    if (varName === 'prototype' || varName === 'constructor') {
        if (target === null || target === void 0 || typeof target === 'function') {
            throw new Error(`[SX] ${name}: Invalid var name ${varName}.`);
        }
    }
    if (target === null || target === void 0 || target === objConstructor) {
        if (objConstructor.hasOwnProperty(varName)) {
            throw new Error(`[SX] ${name}: Invalid var name ${varName}.`);
        }
    }
    if (target === null || target === void 0 || target === funConstructor) {
        // checking 'call', 'arguments', 'caller', ...
        let con = funConstructor;
        while (con) {
            if (con.hasOwnProperty(varName)) {
                throw new Error(`[SX] ${name}: Invalid var name ${varName}.`);
            }
            con = con.__proto__;
        }
    }
    if (typeof target === 'function') {
        if (!target.hasOwnProperty(varName)) {
            // function's prototypes' members
            throw new Error(`[SX] ${name}: Invalid var name ${varName}.`);
        }
    }
    return varName;
}
exports.checkUnsafeVarNamesEx = checkUnsafeVarNamesEx;
//# sourceMappingURL=errors.js.map