"use strict";
// Copyright (c) 2018, Shellyl_N and Authors
// license: ISC
// https://github.com/shellyln
Object.defineProperty(exports, "__esModule", { value: true });
const evaluate_1 = require("../../evaluate");
const errors_1 = require("../../errors");
const core_fn_1 = require("../core/core.fn");
exports.$bitLShift = (state, name) => (...args) => {
    // S expression: (<< number shift)
    //  -> S expr  : number
    errors_1.checkParamsLength('$bitLShift', args, 2, 2);
    let { car, cdr } = core_fn_1.$$firstAndSecond(...args);
    car = evaluate_1.toNumber(car);
    cdr = evaluate_1.toNumber(cdr);
    if (0 <= cdr) {
        return cdr < 32 ? car << cdr : 0;
    }
    else {
        return cdr > -32 ? car >>> (-cdr) : 0;
    }
};
exports.$$bitLShift = exports.$bitLShift(null, null);
exports.$bitSRShift = (state, name) => (...args) => {
    // S expression: (>> number shift)
    //  -> S expr  : number
    errors_1.checkParamsLength('$bitSRShift', args, 2, 2);
    let { car, cdr } = core_fn_1.$$firstAndSecond(...args);
    car = evaluate_1.toNumber(car);
    cdr = evaluate_1.toNumber(cdr);
    if (0 <= cdr) {
        return cdr < 32 ? car >> cdr : (car & 0x080000000) ? -1 : 0;
    }
    else {
        return cdr > -32 ? car << (-cdr) : 0;
    }
};
exports.$$bitSRShift = exports.$bitSRShift(null, null);
exports.$bitURShift = (state, name) => (...args) => {
    // S expression: (>>> number shift)
    //  -> S expr  : number
    errors_1.checkParamsLength('$bitURShift', args, 2, 2);
    let { car, cdr } = core_fn_1.$$firstAndSecond(...args);
    car = evaluate_1.toNumber(car);
    cdr = evaluate_1.toNumber(cdr);
    if (0 <= cdr) {
        return cdr < 32 ? car >>> cdr : 0;
    }
    else {
        return cdr > -32 ? car << (-cdr) : 0;
    }
};
exports.$$bitURShift = exports.$bitURShift(null, null);
exports.$bitNot = (state, name) => (...args) => {
    // S expression: ($bit-not number)
    //  -> S expr  : number
    errors_1.checkParamsLength('$bitNot', args, 1, 1);
    const car = core_fn_1.$$first(...args);
    return ~evaluate_1.toNumber(car);
};
exports.$$bitNot = exports.$bitNot(null, null);
exports.$bitAnd = (state, name) => (...args) => {
    // S expression: ($bit-and numberA numberB)
    //  -> S expr  : number
    errors_1.checkParamsLength('$bitAnd', args, 2);
    const car = core_fn_1.$$first(...args);
    return args.slice(1).reduce((prev, curr) => evaluate_1.toNumber(prev) & evaluate_1.toNumber(curr), evaluate_1.toNumber(car));
};
exports.$$bitAnd = exports.$bitAnd(null, null);
exports.$bitOr = (state, name) => (...args) => {
    // S expression: ($bit-or numberA numberB)
    //  -> S expr  : number
    errors_1.checkParamsLength('$bitOr', args, 2);
    const car = core_fn_1.$$first(...args);
    return args.slice(1).reduce((prev, curr) => evaluate_1.toNumber(prev) | evaluate_1.toNumber(curr), evaluate_1.toNumber(car));
};
exports.$$bitOr = exports.$bitOr(null, null);
exports.$bitXor = (state, name) => (...args) => {
    // S expression: ($bit-xor numberA numberB)
    //  -> S expr  : number
    errors_1.checkParamsLength('$bitXor', args, 2);
    const car = core_fn_1.$$first(...args);
    return args.slice(1).reduce((prev, curr) => evaluate_1.toNumber(prev) ^ evaluate_1.toNumber(curr), evaluate_1.toNumber(car));
};
exports.$$bitXor = exports.$bitXor(null, null);
exports.$add = (state, name) => (...args) => {
    // S expression: (+ number1 ... numberN)
    //  -> S expr  : number
    errors_1.checkParamsLength('$add', args, 1);
    return args.reduce((prev, curr) => evaluate_1.toNumber(prev) + evaluate_1.toNumber(curr), 0);
};
exports.$$add = exports.$add(null, null);
exports.$sub = (state, name) => (...args) => {
    // S expression: (- number1 ... numberN)
    //  -> S expr  : number
    errors_1.checkParamsLength('$sub', args, 1);
    const car = core_fn_1.$$first(...args);
    const last = args.slice(1);
    if (last.length === 0) {
        // negate
        return -evaluate_1.toNumber(car);
    }
    else {
        // subtract
        return args.slice(1).reduce((prev, curr) => evaluate_1.toNumber(prev) - evaluate_1.toNumber(curr), evaluate_1.toNumber(car));
    }
};
exports.$$sub = exports.$sub(null, null);
exports.$mul = (state, name) => (...args) => {
    // S expression: (* number1 ... numberN)
    //  -> S expr  : number
    errors_1.checkParamsLength('$mul', args, 2);
    const car = core_fn_1.$$first(...args);
    return args.slice(1).reduce((prev, curr) => evaluate_1.toNumber(prev) * evaluate_1.toNumber(curr), evaluate_1.toNumber(car));
};
exports.$$mul = exports.$mul(null, null);
exports.$sup = (state, name) => (...args) => {
    // S expression: (** number1 ... numberN)
    //  -> S expr  : number
    errors_1.checkParamsLength('$sup', args, 2);
    const car = core_fn_1.$$first(...args);
    return args.slice(1).reduce((prev, curr) => Math.pow(evaluate_1.toNumber(prev), evaluate_1.toNumber(curr)), evaluate_1.toNumber(car));
};
exports.$$sup = exports.$sup(null, null);
exports.$div = (state, name) => (...args) => {
    // S expression: (/ number1 ... numberN)
    //  -> S expr  : number
    errors_1.checkParamsLength('$div', args, 2);
    const car = core_fn_1.$$first(...args);
    return args.slice(1).reduce((prev, curr) => evaluate_1.toNumber(prev) / evaluate_1.toNumber(curr), evaluate_1.toNumber(car));
};
exports.$$div = exports.$div(null, null);
exports.$mod = (state, name) => (...args) => {
    // S expression: (% number1 ... numberN)
    //  -> S expr  : number
    errors_1.checkParamsLength('$mod', args, 2);
    const car = core_fn_1.$$first(...args);
    return args.slice(1).reduce((prev, curr) => evaluate_1.toNumber(prev) % evaluate_1.toNumber(curr), evaluate_1.toNumber(car));
};
exports.$$mod = exports.$mod(null, null);
exports.$max = (state, name) => (...args) => {
    // S expression: ($max val1 ... valN)
    //  -> S expr  : value
    return Math.max(...(args.map(x => evaluate_1.toNumber(x))));
};
exports.$$max = exports.$max(null, null);
exports.$min = (state, name) => (...args) => {
    // S expression: ($min val1 ... valN)
    //  -> S expr  : value
    return Math.min(...(args.map(x => evaluate_1.toNumber(x))));
};
exports.$$min = exports.$min(null, null);
exports.$avg = (state, name) => (...args) => {
    // S expression: ($avg val1 ... valN)
    //  -> S expr  : value
    const a = args.map(x => evaluate_1.toNumber(x));
    return a.length > 0 ? a.reduce((prev, curr) => prev + curr, 0) / a.length : NaN;
};
exports.$$avg = exports.$avg(null, null);
exports.$floor = (state, name) => (...args) => {
    // S expression: ($floor number)
    //  -> S expr  : number
    errors_1.checkParamsLength('$floor', args, 1, 1);
    return Math.floor(evaluate_1.toNumber(core_fn_1.$$first(...args)));
};
exports.$$floor = exports.$floor(null, null);
exports.$ceil = (state, name) => (...args) => {
    // S expression: ($ceil number)
    //  -> S expr  : number
    errors_1.checkParamsLength('$ceil', args, 1, 1);
    return Math.ceil(evaluate_1.toNumber(core_fn_1.$$first(...args)));
};
exports.$$ceil = exports.$ceil(null, null);
exports.$round = (state, name) => (...args) => {
    // S expression: ($round number)
    //  -> S expr  : number
    errors_1.checkParamsLength('$round', args, 1, 1);
    return Math.round(evaluate_1.toNumber(core_fn_1.$$first(...args)));
};
exports.$$round = exports.$round(null, null);
exports.$abs = (state, name) => (...args) => {
    // S expression: ($abs number)
    //  -> S expr  : number
    errors_1.checkParamsLength('$abs', args, 1, 1);
    return Math.abs(evaluate_1.toNumber(core_fn_1.$$first(...args)));
};
exports.$$abs = exports.$abs(null, null);
exports.$sign = (state, name) => (...args) => {
    // S expression: ($sign number)
    //  -> S expr  : number
    errors_1.checkParamsLength('$sign', args, 1, 1);
    return Math.sign(evaluate_1.toNumber(core_fn_1.$$first(...args)));
};
exports.$$sign = exports.$sign(null, null);
//# sourceMappingURL=arithmetic.fn.js.map