"use strict";
// Copyright (c) 2018, Shellyl_N and Authors
// license: ISC
// https://github.com/shellyln
Object.defineProperty(exports, "__esModule", { value: true });
const errors_1 = require("../../errors");
exports.macros = [{
        name: '$incl',
        fn: (state, name) => (list) => {
            // S expression: ($incl name)
            //  -> S expr  : ($set name ($add name 1))
            errors_1.checkParamsLength('$incl', list, 2, 2);
            return [{ symbol: '$set' },
                list[1],
                [{ symbol: '$add' }, list[1], 1],
            ];
        },
    }, {
        name: '++',
        fn: (state, name) => (list) => {
            // S expression: (++ name)
            //  -> S expr  : ($set name ($add name 1))
            errors_1.checkParamsLength('++', list, 2, 2);
            return [{ symbol: '$set' },
                list[1],
                [{ symbol: '$add' }, list[1], 1],
            ];
        },
    }, {
        name: '$decl',
        fn: (state, name) => (list) => {
            // S expression: ($decl name)
            //  -> S expr  : ($set name ($add name -1))
            errors_1.checkParamsLength('$decl', list, 2, 2);
            return [{ symbol: '$set' },
                list[1],
                [{ symbol: '$add' }, list[1], -1],
            ];
        },
    }, {
        name: '--',
        fn: (state, name) => (list) => {
            // S expression: (-- name)
            //  -> S expr  : ($set name ($add name -1))
            errors_1.checkParamsLength('--', list, 2, 2);
            return [{ symbol: '$set' },
                list[1],
                [{ symbol: '$add' }, list[1], -1],
            ];
        },
    }, {
        name: '$incln',
        fn: (state, name) => (list) => {
            // S expression: ($incln name v)
            //  -> S expr  : ($set name ($add name v))
            errors_1.checkParamsLength('$incln', list, 3, 3);
            if (typeof list[2] !== 'number') {
                throw new Error(`[SX] $incln: Invalid parameter: arg(1) is not number.`);
            }
            return [{ symbol: '$set' },
                list[1],
                [{ symbol: '$add' }, list[1], list[2]],
            ];
        },
    }, {
        name: '+=',
        fn: (state, name) => (list) => {
            // S expression: (++ name v)
            //  -> S expr  : ($set name ($add name v))
            errors_1.checkParamsLength('+=', list, 3, 3);
            if (typeof list[2] !== 'number') {
                throw new Error(`[SX] +=: Invalid parameter: arg(1) is not number.`);
            }
            return [{ symbol: '$set' },
                list[1],
                [{ symbol: '$add' }, list[1], list[2]],
            ];
        },
    }, {
        name: '$decln',
        fn: (state, name) => (list) => {
            // S expression: ($decln name v)
            //  -> S expr  : ($set name ($add name -v))
            errors_1.checkParamsLength('$decln', list, 3, 3);
            if (typeof list[2] !== 'number') {
                throw new Error(`[SX] $decln: Invalid parameter: arg(1) is not number.`);
            }
            return [{ symbol: '$set' },
                list[1],
                [{ symbol: '$add' }, list[1], -list[2]],
            ];
        },
    }, {
        name: '-=',
        fn: (state, name) => (list) => {
            // S expression: (-- name)
            //  -> S expr  : ($set name ($add name -v))
            errors_1.checkParamsLength('-=', list, 3, 3);
            if (typeof list[2] !== 'number') {
                throw new Error(`[SX] -=: Invalid parameter: arg(1) is not number.`);
            }
            return [{ symbol: '$set' },
                list[1],
                [{ symbol: '$add' }, list[1], -list[2]],
            ];
        },
    }];
exports.default = exports.macros;
//# sourceMappingURL=arithmetic.macro.js.map