"use strict";
// Copyright (c) 2018, Shellyl_N and Authors
// license: ISC
// https://github.com/shellyln
Object.defineProperty(exports, "__esModule", { value: true });
const errors_1 = require("../../errors");
const core_fn_1 = require("../core/core.fn");
// tslint:disable-next-line:variable-name
exports.$__letAsync = (state, name) => (...args) => {
    // S expression: ($__let-async 'nameStrOrSymbol promise)
    //  -> S expr  : promise
    errors_1.checkParamsLength('$__letAsync', args, 2, 2);
    let promise = args[1];
    if (typeof promise !== 'object' || typeof promise.then !== 'function') {
        promise = Promise.resolve(promise);
    }
    promise = promise.then(v => {
        try {
            core_fn_1.$__let(state, '')(args[0], v);
            return v;
        }
        catch (e) {
            return Promise.reject(e);
        }
    });
    return promise;
};
// tslint:disable-next-line:variable-name
exports.$$__letAsync = exports.$__letAsync(null, null);
// tslint:disable-next-line:variable-name
exports.$__setAsync = (state, name) => (...args) => {
    // S expression: ($__set-async 'nameOrListOfNameOrIndex promise)
    //  -> S expr  : promise
    errors_1.checkParamsLength('$__setAsync', args, 2, 2);
    let promise = args[1];
    if (typeof promise !== 'object' || typeof promise.then !== 'function') {
        promise = Promise.resolve(promise);
    }
    promise = promise.then(v => {
        try {
            core_fn_1.$__set(state, '')(args[0], v);
            return v;
        }
        catch (e) {
            return Promise.reject(e);
        }
    });
    return promise;
};
// tslint:disable-next-line:variable-name
exports.$$__setAsync = exports.$__setAsync(null, null);
exports.$then = (state, name) => (...args) => {
    // S expression: ($then promise (lambda (val) ...) (lambda (err) ...))
    //  -> S expr  : promise
    errors_1.checkParamsLength('$then', args, 2, 3);
    let promise = args[0];
    if (typeof promise !== 'object' || typeof promise.then !== 'function') {
        promise = Promise.resolve(promise);
    }
    if (typeof args[2] === 'function') {
        promise = promise.then(args[1], args[2]);
    }
    else {
        if (typeof args[1] !== 'function') {
            throw new Error(`[SX] $then: Invalid argument(s): args[1] is not function.`);
        }
        promise = promise.then(args[1]);
    }
    return promise;
};
exports.$$then = exports.$then(null, null);
exports.$resolveAll = (state, name) => (...args) => {
    // S expression: ($resolve-all promise1 ... promiseN)
    //  -> S expr  : promise
    const promises = args.slice(0);
    for (let i = 0; i < promises.length; i++) {
        if (typeof promises[i] !== 'object' || typeof promises[i].then !== 'function') {
            promises[i] = Promise.resolve(promises[i]);
        }
    }
    return Promise.all(promises);
};
exports.$$resolveAll = exports.$resolveAll(null, null);
exports.$resolveAny = (state, name) => (...args) => {
    // S expression: ($resolve-any promise1 ... promiseN)
    //  -> S expr  : promise
    const promises = args.slice(0);
    for (let i = 0; i < promises.length; i++) {
        if (typeof promises[i] !== 'object' || typeof promises[i].then !== 'function') {
            promises[i] = Promise.resolve(promises[i]);
        }
    }
    // https://stackoverflow.com/questions/39940152/get-first-fulfilled-promise
    // firstOf: This will return the value of the first fulfilled promise,
    //          or if all reject, an array of rejection reasons.
    const invert = (p) => new Promise((res, rej) => p.then(rej, res));
    const firstOf = (ps) => invert(Promise.all(ps.map(invert)));
    return firstOf(promises);
};
exports.$$resolveAny = exports.$resolveAny(null, null);
exports.$resolvePipe = (state, name) => (...args) => {
    // S expression: ($resolve-pipe promise<val1> (lambda (val1) ... promiseOrVal2) (lambda (val2) ... promiseOrVal3) ... (lambda (valN-1) ... promiseOrValN))
    //  -> S expr  : promise
    // remarks: If the formal argument lambda is a non-lambda value, the value is then piped as is.
    errors_1.checkParamsLength('$resolvePipe', args, 1);
    let promise = args[0];
    if (typeof promise !== 'object' || typeof promise.then !== 'function') {
        promise = Promise.resolve(promise);
    }
    const lambdas = args.slice(1);
    for (let i = 0; i < lambdas.length; i++) {
        if (typeof lambdas[i] !== 'function') {
            const v = lambdas[i];
            lambdas[i] = () => v;
        }
    }
    let p = promise;
    for (const l of lambdas) {
        p = p.then(l);
    }
    return p;
};
exports.$$resolvePipe = exports.$resolvePipe(null, null);
exports.$resolveFork = (state, name) => (...args) => {
    // S expression: ($resolve-fork promise<val1> (lambda (val1) ... promiseOrVal2a) ... (lambda (val1) ... promiseOrVal2z))
    //  -> S expr  : (promise<val2a> ... promise<val2z>)
    // remarks: If the formal argument lambda is a non-lambda value, the value is then piped as is.
    errors_1.checkParamsLength('$resolveFork', args, 1);
    let promise = args[0];
    if (typeof promise !== 'object' || typeof promise.then !== 'function') {
        promise = Promise.resolve(promise);
    }
    const lambdas = args.slice(1);
    for (let i = 0; i < lambdas.length; i++) {
        if (typeof lambdas[i] !== 'function') {
            const v = lambdas[i];
            lambdas[i] = () => v;
        }
    }
    const resolvers = new Array(lambdas.length);
    const rejectors = new Array(lambdas.length);
    const pa = [];
    for (let i = 0; i < lambdas.length; i++) {
        pa.push(new Promise((resolve, reject) => {
            resolvers[i] = (v) => {
                let lp = lambdas[i](v);
                if (typeof lp !== 'object' || typeof lp.then !== 'function') {
                    lp = Promise.resolve(lp);
                }
                lp
                    .then(x => resolve(x))
                    .catch(e => reject(e));
            };
            rejectors[i] = reject;
        }));
    }
    promise.then(v => resolvers.forEach(f => f(v)), e => rejectors.forEach(f => f(e)));
    return pa;
};
exports.$$resolveFork = exports.$resolveFork(null, null);
//# sourceMappingURL=concurrent.fn.js.map