// Copyright (c) 2018, Shellyl_N and Authors
// license: ISC
// https://github.com/shellyln
import { isSymbol } from '../ast';
import { resolveSplice, resolveMacro, resolveValueSymbol, getScope } from '../evaluate';
import { setEvaluationCount, checkUnsafeVarNames } from '../errors';
export function applyMacros(state, tok) {
    let r = tok;
    for (;;) {
        if (Array.isArray(r)) {
            if (r.length === 0) {
                break;
            }
            r = resolveSplice(state, r);
            const sym = isSymbol(r[0]);
            if (sym) {
                const m = resolveMacro(state, sym, r);
                if (m) {
                    r = m.fn(m.actualArgs);
                }
                else {
                    break;
                }
            }
            else {
                break;
            }
        }
        else {
            break;
        }
        setEvaluationCount(state);
    }
    return r;
}
export function stripQuote(state, tok) {
    if (!(Array.isArray(tok) && isSymbol(tok[0], state.config.reservedNames.quote))) {
        throw new Error(`[SX] stripQuote: token is not quoted.`);
    }
    return tok[1];
}
export function stripQuoteOrPass(state, tok) {
    if (Array.isArray(tok) && isSymbol(tok[0], state.config.reservedNames.quote)) {
        return tok[1];
    }
    else {
        return tok;
    }
}
export function getScope_stateApplied(state) {
    return (function () { return getScope(state); });
}
export function resolveValueSymbol_dynamic(state, varName) {
    checkUnsafeVarNames('(compiler)resolveValueSymbol_dynamic', varName);
    return (function () { return resolveValueSymbol(state, { symbol: varName }); });
}
export function checkUnsafeVarNames_dynamic(name) {
    return (function (varName) { return checkUnsafeVarNames(name, varName); });
}
//# sourceMappingURL=helpers.js.map