// Copyright (c) 2018, Shellyl_N and Authors
// license: ISC
// https://github.com/shellyln
import { toNumber } from '../../evaluate';
import { checkParamsLength } from '../../errors';
import { $$first, $$firstAndSecond } from '../core/core.fn';
export const $bitLShift = (state, name) => (...args) => {
    // S expression: (<< number shift)
    //  -> S expr  : number
    checkParamsLength('$bitLShift', args, 2, 2);
    let { car, cdr } = $$firstAndSecond(...args);
    car = toNumber(car);
    cdr = toNumber(cdr);
    if (0 <= cdr) {
        return cdr < 32 ? car << cdr : 0;
    }
    else {
        return cdr > -32 ? car >>> (-cdr) : 0;
    }
};
export const $$bitLShift = $bitLShift(null, null);
export const $bitSRShift = (state, name) => (...args) => {
    // S expression: (>> number shift)
    //  -> S expr  : number
    checkParamsLength('$bitSRShift', args, 2, 2);
    let { car, cdr } = $$firstAndSecond(...args);
    car = toNumber(car);
    cdr = toNumber(cdr);
    if (0 <= cdr) {
        return cdr < 32 ? car >> cdr : (car & 0x080000000) ? -1 : 0;
    }
    else {
        return cdr > -32 ? car << (-cdr) : 0;
    }
};
export const $$bitSRShift = $bitSRShift(null, null);
export const $bitURShift = (state, name) => (...args) => {
    // S expression: (>>> number shift)
    //  -> S expr  : number
    checkParamsLength('$bitURShift', args, 2, 2);
    let { car, cdr } = $$firstAndSecond(...args);
    car = toNumber(car);
    cdr = toNumber(cdr);
    if (0 <= cdr) {
        return cdr < 32 ? car >>> cdr : 0;
    }
    else {
        return cdr > -32 ? car << (-cdr) : 0;
    }
};
export const $$bitURShift = $bitURShift(null, null);
export const $bitNot = (state, name) => (...args) => {
    // S expression: ($bit-not number)
    //  -> S expr  : number
    checkParamsLength('$bitNot', args, 1, 1);
    const car = $$first(...args);
    return ~toNumber(car);
};
export const $$bitNot = $bitNot(null, null);
export const $bitAnd = (state, name) => (...args) => {
    // S expression: ($bit-and numberA numberB)
    //  -> S expr  : number
    checkParamsLength('$bitAnd', args, 2);
    const car = $$first(...args);
    return args.slice(1).reduce((prev, curr) => toNumber(prev) & toNumber(curr), toNumber(car));
};
export const $$bitAnd = $bitAnd(null, null);
export const $bitOr = (state, name) => (...args) => {
    // S expression: ($bit-or numberA numberB)
    //  -> S expr  : number
    checkParamsLength('$bitOr', args, 2);
    const car = $$first(...args);
    return args.slice(1).reduce((prev, curr) => toNumber(prev) | toNumber(curr), toNumber(car));
};
export const $$bitOr = $bitOr(null, null);
export const $bitXor = (state, name) => (...args) => {
    // S expression: ($bit-xor numberA numberB)
    //  -> S expr  : number
    checkParamsLength('$bitXor', args, 2);
    const car = $$first(...args);
    return args.slice(1).reduce((prev, curr) => toNumber(prev) ^ toNumber(curr), toNumber(car));
};
export const $$bitXor = $bitXor(null, null);
export const $add = (state, name) => (...args) => {
    // S expression: (+ number1 ... numberN)
    //  -> S expr  : number
    checkParamsLength('$add', args, 1);
    return args.reduce((prev, curr) => toNumber(prev) + toNumber(curr), 0);
};
export const $$add = $add(null, null);
export const $sub = (state, name) => (...args) => {
    // S expression: (- number1 ... numberN)
    //  -> S expr  : number
    checkParamsLength('$sub', args, 1);
    const car = $$first(...args);
    const last = args.slice(1);
    if (last.length === 0) {
        // negate
        return -toNumber(car);
    }
    else {
        // subtract
        return args.slice(1).reduce((prev, curr) => toNumber(prev) - toNumber(curr), toNumber(car));
    }
};
export const $$sub = $sub(null, null);
export const $mul = (state, name) => (...args) => {
    // S expression: (* number1 ... numberN)
    //  -> S expr  : number
    checkParamsLength('$mul', args, 2);
    const car = $$first(...args);
    return args.slice(1).reduce((prev, curr) => toNumber(prev) * toNumber(curr), toNumber(car));
};
export const $$mul = $mul(null, null);
export const $sup = (state, name) => (...args) => {
    // S expression: (** number1 ... numberN)
    //  -> S expr  : number
    checkParamsLength('$sup', args, 2);
    const car = $$first(...args);
    return args.slice(1).reduce((prev, curr) => Math.pow(toNumber(prev), toNumber(curr)), toNumber(car));
};
export const $$sup = $sup(null, null);
export const $div = (state, name) => (...args) => {
    // S expression: (/ number1 ... numberN)
    //  -> S expr  : number
    checkParamsLength('$div', args, 2);
    const car = $$first(...args);
    return args.slice(1).reduce((prev, curr) => toNumber(prev) / toNumber(curr), toNumber(car));
};
export const $$div = $div(null, null);
export const $mod = (state, name) => (...args) => {
    // S expression: (% number1 ... numberN)
    //  -> S expr  : number
    checkParamsLength('$mod', args, 2);
    const car = $$first(...args);
    return args.slice(1).reduce((prev, curr) => toNumber(prev) % toNumber(curr), toNumber(car));
};
export const $$mod = $mod(null, null);
export const $max = (state, name) => (...args) => {
    // S expression: ($max val1 ... valN)
    //  -> S expr  : value
    return Math.max(...(args.map(x => toNumber(x))));
};
export const $$max = $max(null, null);
export const $min = (state, name) => (...args) => {
    // S expression: ($min val1 ... valN)
    //  -> S expr  : value
    return Math.min(...(args.map(x => toNumber(x))));
};
export const $$min = $min(null, null);
export const $avg = (state, name) => (...args) => {
    // S expression: ($avg val1 ... valN)
    //  -> S expr  : value
    const a = args.map(x => toNumber(x));
    return a.length > 0 ? a.reduce((prev, curr) => prev + curr, 0) / a.length : NaN;
};
export const $$avg = $avg(null, null);
export const $floor = (state, name) => (...args) => {
    // S expression: ($floor number)
    //  -> S expr  : number
    checkParamsLength('$floor', args, 1, 1);
    return Math.floor(toNumber($$first(...args)));
};
export const $$floor = $floor(null, null);
export const $ceil = (state, name) => (...args) => {
    // S expression: ($ceil number)
    //  -> S expr  : number
    checkParamsLength('$ceil', args, 1, 1);
    return Math.ceil(toNumber($$first(...args)));
};
export const $$ceil = $ceil(null, null);
export const $round = (state, name) => (...args) => {
    // S expression: ($round number)
    //  -> S expr  : number
    checkParamsLength('$round', args, 1, 1);
    return Math.round(toNumber($$first(...args)));
};
export const $$round = $round(null, null);
export const $abs = (state, name) => (...args) => {
    // S expression: ($abs number)
    //  -> S expr  : number
    checkParamsLength('$abs', args, 1, 1);
    return Math.abs(toNumber($$first(...args)));
};
export const $$abs = $abs(null, null);
export const $sign = (state, name) => (...args) => {
    // S expression: ($sign number)
    //  -> S expr  : number
    checkParamsLength('$sign', args, 1, 1);
    return Math.sign(toNumber($$first(...args)));
};
export const $$sign = $sign(null, null);
//# sourceMappingURL=arithmetic.fn.js.map